<?php

namespace Drupal\simple_sitemap_arbitrary_links;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\ReplaceCommand;

/**
 * Provides form alterations for custom sitemap links.
 */
class CustomSitemapLinksFormAlter {

  /**
   * Alters the custom links form to add a table of editable links.
   *
   * @param array &$form
   *   The form array to modify.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \RuntimeException
   *   If database query or language retrieval fails.
   */
  public static function alterForm(array &$form, FormStateInterface $form_state) {
    $db = \Drupal::database();
    try {
      $links = $db->select('custom_sitemap_links', 'c')
        ->fields(
                'c', [
                  'id',
                  'url',
                  'priority',
                  'changefreq',
                  'lastmod',
                  'langcode',
                ]
            )
        ->orderBy('id')
        ->execute()
        ->fetchAllAssoc('id');
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to load custom sitemap links: ' . $e->getMessage());
    }

    $form['custom_links'] = [
      '#type' => 'fieldset',
      '#title' => t('Custom Arbitrary Links'),
      '#tree' => TRUE,
      '#description' => t('Define custom URLs for the XML sitemap, independent of Drupal entities.'),
      '#prefix' => '<div id="custom-links-wrapper">',
      '#suffix' => '</div>',
    ];

    $form['custom_links']['links'] = [
      '#type' => 'table',
      '#header' => [
        t('URL'),
        t('Priority'),
        t('Change Frequency'),
        t('Last Modified'),
        t('Language'),
        t('Operations'),
      ],
    ];

    try {
      $languages = \Drupal::languageManager()->getLanguages();
      $lang_options = array_map(fn($lang) => $lang->getName(), $languages);
      $default_lang = \Drupal::languageManager()->getDefaultLanguage()->getId();
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to retrieve language options: ' . $e->getMessage());
    }

    foreach ($links as $id => $link) {
      $form['custom_links']['links'][$id] = [
        'url' => [
          '#type' => 'textfield',
          '#default_value' => $link->url,
          '#size' => 40,
        ],
        'priority' => [
          '#type' => 'select',
          '#default_value' => $link->priority ?? '1.0',
          '#options' => array_combine(
            $p = [
              '0.1',
              '0.2',
              '0.3',
              '0.4',
              '0.5',
              '0.6',
              '0.7',
              '0.8',
              '0.9',
              '1.0',
            ], $p
          ),
        ],
        'changefreq' => [
          '#type' => 'select',
          '#default_value' => $link->changefreq ?? '',
          '#options' => ['' => t('- Not specified -')] + array_map(
            't', [
              'always',
              'hourly',
              'daily',
              'weekly',
              'monthly',
              'yearly',
              'never',
            ]
          ),
        ],
        'lastmod' => [
          '#type' => 'date',
          '#default_value' => $link->lastmod ? date('Y-m-d', strtotime($link->lastmod)) : '',
        ],
        'langcode' => [
          '#type' => 'select',
          '#default_value' => $link->langcode ?? $default_lang,
          '#options' => $lang_options,
        ],
        'remove' => [
          '#type' => 'submit',
          '#value' => t('Remove'),
          '#name' => "remove_$id",
          '#delta' => $id,
          '#submit' => [[self::class, 'removeLink']],
        ],
      ];
    }

    $form['custom_links']['add_link'] = [
      '#type' => 'submit',
      '#value' => t('Add new link'),
      '#ajax' => [
        'callback' => [self::class, 'addLinkAjax'],
        'wrapper' => 'custom-links-wrapper',
        'event' => 'click',
      ],
    ];

    if ($form_state->get('scroll_to_custom_links')) {
      $form['#attached']['library'][] = 'simple_sitemap_arbitrary_links/scroll_to_custom_links';
    }

    $form['#submit'][] = [self::class, 'submitForm'];
  }

  /**
   * AJAX callback to add a new link to the database and rebuild the form.
   *
   * @param array &$form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   The AJAX response to update the form.
   *
   * @throws \RuntimeException
   *   If the database insert or form rebuild fails.
   */
  public static function addLinkAjax(array &$form, FormStateInterface $form_state) {
    $db = \Drupal::database();
    $now = (new \DateTime('now', new \DateTimeZone('UTC')))->format('c');

    try {
      $new_id = $db->insert('custom_sitemap_links')
        ->fields(
                [
                  'url' => '',
                  'priority' => '1.0',
                  'changefreq' => '',
                  'lastmod' => $now,
                  'langcode' => \Drupal::languageManager()
                    ->getDefaultLanguage()
                    ->getId(),
                ]
            )
        ->execute();
      if (!$new_id) {
        throw new \RuntimeException('Failed to insert new sitemap link.');
      }
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Error adding new sitemap link: ' . $e->getMessage());
    }

    $form_state->setRebuild(TRUE);
    try {
      $rebuilt_form = \Drupal::formBuilder()
        ->rebuildForm(
                $form_state->getFormObject()
                  ->getFormId(), $form_state, $form
            );
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to rebuild form after adding link: ' . $e->getMessage());
    }

    return (new AjaxResponse())->addCommand(new ReplaceCommand('#custom-links-wrapper', $rebuilt_form['custom_links']));
  }

  /**
   * Submit handler to remove a link from the database.
   *
   * @param array &$form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \RuntimeException
   *   If the link ID is invalid or deletion fails.
   */
  public static function removeLink(array &$form, FormStateInterface $form_state) {
    $trigger = $form_state->getTriggeringElement();
    if (!isset($trigger['#delta']) || !is_numeric($trigger['#delta'])) {
      throw new \RuntimeException('Invalid link ID for removal.');
    }
    $id = $trigger['#delta'];

    try {
      $deleted = \Drupal::database()->delete('custom_sitemap_links')
        ->condition('id', $id)
        ->execute();
      if ($deleted === 0) {
        throw new \RuntimeException("No link found with ID $id to remove.");
      }
    }
    catch (\Exception $e) {
      throw new \RuntimeException('Failed to remove sitemap link: ' . $e->getMessage());
    }

    $form_state->setValues([]);
    $form_state->setUserInput([]);
    $form_state->set('scroll_to_custom_links', TRUE);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Submit handler to save form changes and regenerate the sitemap.
   *
   * @param array &$form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   *
   * @throws \RuntimeException
   *   If database updates or sitemap regeneration fails.
   */
  public static function submitForm(array &$form, FormStateInterface $form_state) {
    $db = \Drupal::database();
    $base_url = \Drupal::request()->getSchemeAndHttpHost();
    $values = $form_state->getValue(['custom_links', 'links']) ?: [];

    foreach ($values as $id => $link) {
      if (!empty($link['url'])) {
        $url = $link['url'];
        $url = (strpos($url, '/') === 0 || preg_match('#^https?://#', $url)) ? $url : "/$url";
        $url = (strpos($url, '/') === 0) ? $base_url . $url : $url;

        try {
          $lastmod = !empty($link['lastmod']) ? (new \DateTime($link['lastmod'], new \DateTimeZone('UTC')))->format('c') : NULL;
        }
        catch (\Exception $e) {
          throw new \RuntimeException('Invalid last modified date for link ID ' . $id . ': ' . $e->getMessage());
        }

        try {
          $db->merge('custom_sitemap_links')
            ->key('id', (string) $id)
            ->fields(
                    [
                      'url' => $url,
                      'priority' => $link['priority'],
                      'changefreq' => $link['changefreq'],
                      'lastmod' => $lastmod,
                      'langcode' => $link['langcode'],
                    ]
                )
            ->execute();
        }
        catch (\Exception $e) {
          throw new \RuntimeException('Failed to save sitemap link ID ' . $id . ': ' . $e->getMessage());
        }
      }
    }

    if ($form_state->getValue('simple_sitemap_regenerate_now')) {
      try {
        \Drupal::service('simple_sitemap.generator')->generate();
      }
      catch (\Exception $e) {
        throw new \RuntimeException('Failed to regenerate sitemap: ' . $e->getMessage());
      }
    }
  }

}
